﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using HeuristicLab.Common;

namespace HeuristicLab.Problems.DataAnalysis {
  public class OnlineWeightedResidualsMeanSquaredErrorCalculator : IOnlineCalculator {

    private double sse;
    private int n;
    public double WeightedResidualsMeanSquaredError {
      get {
        return n > 0 ? sse / n : 0.0;
      }
    }

    public double PositiveClassValue { get; private set; }
    public double ClassValueMax { get; private set; }
    public double ClassValueMin { get; private set; }
    public double ObviousResidualsWeight { get; private set; }
    public double PositiveClassResidualsWeight { get; private set; }
    public double NegativeClassResidualsWeight { get; private set; }

    public OnlineWeightedResidualsMeanSquaredErrorCalculator(double positiveClassValue, double classValueMax, double classValueMin,
                                                                double obviousResidualsWeight, double positiveClassResidualsWeight, double negativeClassResidualsWeight) {
      PositiveClassValue = positiveClassValue;
      ClassValueMax = classValueMax;
      ClassValueMin = classValueMin;
      ObviousResidualsWeight = obviousResidualsWeight;
      PositiveClassResidualsWeight = positiveClassResidualsWeight;
      NegativeClassResidualsWeight = negativeClassResidualsWeight;
      Reset();
    }

    #region IOnlineCalculator Members
    private OnlineCalculatorError errorState;
    public OnlineCalculatorError ErrorState {
      get { return errorState; }
    }
    public double Value {
      get { return WeightedResidualsMeanSquaredError; }
    }
    public void Reset() {
      n = 0;
      sse = 0.0;
      errorState = OnlineCalculatorError.InsufficientElementsAdded;
    }

    public void Add(double original, double estimated) {
      if (double.IsNaN(estimated) || double.IsInfinity(estimated) ||
          double.IsNaN(original) || double.IsInfinity(original) || (errorState & OnlineCalculatorError.InvalidValueAdded) > 0) {
        errorState = errorState | OnlineCalculatorError.InvalidValueAdded;
      } else {
        double error = estimated - original;
        double weight;
        //apply weight
        if (estimated > ClassValueMax || estimated < ClassValueMin) {
          weight = ObviousResidualsWeight;
        } else if (original.IsAlmost(PositiveClassValue))
        {
          weight = PositiveClassResidualsWeight;
        } else
        {
          weight = NegativeClassResidualsWeight;
        }
        error *= weight;
        sse += error * error;
        n++;
        errorState = errorState & (~OnlineCalculatorError.InsufficientElementsAdded);        // n >= 1
      }
    }
    #endregion

    public static double Calculate(IEnumerable<double> originalValues, IEnumerable<double> estimatedValues, double positiveClassValue, double classValueMax, double classValueMin,
                                                                double obviousValuesWeight, double positiveClassValuesWeight, double negativeClassValuesWeight, out OnlineCalculatorError errorState) {
      IEnumerator<double> originalEnumerator = originalValues.GetEnumerator();
      IEnumerator<double> estimatedEnumerator = estimatedValues.GetEnumerator();
      OnlineWeightedResidualsMeanSquaredErrorCalculator calculator = new OnlineWeightedResidualsMeanSquaredErrorCalculator(positiveClassValue, classValueMax, classValueMin, obviousValuesWeight, positiveClassValuesWeight, negativeClassValuesWeight);

      // always move forward both enumerators (do not use short-circuit evaluation!)
      while (originalEnumerator.MoveNext() & estimatedEnumerator.MoveNext()) {
        double original = originalEnumerator.Current;
        double estimated = estimatedEnumerator.Current;
        calculator.Add(original, estimated);
        if (calculator.ErrorState != OnlineCalculatorError.None) break;
      }

      // check if both enumerators are at the end to make sure both enumerations have the same length
      if (calculator.ErrorState == OnlineCalculatorError.None &&
         (estimatedEnumerator.MoveNext() || originalEnumerator.MoveNext())) {
        throw new ArgumentException("Number of elements in originalValues and estimatedValues enumerations doesn't match.");
      } else {
        errorState = calculator.ErrorState;
        return calculator.WeightedResidualsMeanSquaredError;
      }
    }
  }
}
